
import re
import json

from typing import Union, Any, Optional

class DType:
    def __init__(self, name: str, value: str, default: Any = None,
        has_numeric_value: bool = False):
        """A datatype for ToyDB.

        :param name: Name of the datatype
        :param value: Format string corresponding to python's struct
            `format characters <https://docs.python.org/3/library/struct.html#format-characters>`_.
        :param default: Default value of the datatype
        :param has_numeric_value: Is there a number in the ``value``
            parameter (especially useful with strings).
        """
        self.name = name
        self.value = value
        self.default = default
        self.has_numeric_value = has_numeric_value

    def __repr__(self):
        return f"<dtype: {self.name}>"

    def __str__(self):
        return str(self.value)

    def __call__(self, n: Optional[int] = None) -> str:
        """Get ``DType``'s format string with optional
        repeat number prepended.

        :param n: Prefix int prepended to ``self.value``.
            With strings, it represents the max length of
            the string. If ``n`` is ``None``, just the
            value is returned. If ``DType`` already has
            a number in its ``self.value`` string, ``n``
            must be set to ``None``.
        :return: ``struct.Struct`` format string for ``DType``.
        """
        if n is None: n = ""
        else:
            assert n > 0
            assert not self.has_numeric_value
            n = str(n)
        return f"{n}{self.value}"

    def __getitem__(self, n: int) -> "DType":
        """Another way of calling self.subtype()

        :param n: Prefix int prepended to ``self.value``.
        :return: Subtype ``DType`` with prepended count.
        """
        return self.subtype(n)

    def getLength(self):
        n = "".join(c for c in self.value if c.isdigit())
        if len(n) == 0: return None
        return int(n)

    def validate(self, val) -> bool:
        """Validates the input's type.

        :param val: Value being
        :returns: Is ``val`` a valid instance of this dtype?
        """
        if val is None: return True
        if not isinstance(val,type(self.default)):
            return False
        if isinstance(val,str):
            if len(val) > self.getLength():
                return False
        return True

    def subtype(self, n: int):
        """Creates a new DType with the same format character,
        but with a prefix number. Useful when creating

        :param n: Prefix number prepended to ``self.value``.
            With strings, it represents the max length of
            the string.
        """
        return self.__class__(
            f"{self.name}[{n}]",
            f"{n}{self.value}",
            self.default,
            True)

class JSONEncoder(json.JSONEncoder):
    """Subclass of `json.JSONEncoder` that is able
    to encode `dtype.DType` when calling `json.dump()`.
    """

    def default(self, obj: dict):
        """Handles decoding string data types as
        ``DType`` objects.
        """
        if isinstance(obj,DType):
            return obj.value
        else:
            return super().default(obj)


I32 = DType("Int32","i",0)
I64 = DType("Int64","l",0)
F32 = DType("Float32","f",0)
F64 = DType("Float64","d",0)
BOOL = DType("Bool","?",False)
CHAR = DType("Char","c","")
STRING = DType("String","s","")
STRING50 = STRING.subtype(50)

valid_chars = "ilfd?cs"
supported_types = [
    I32,
    I64,
    F32,
    F64,
    BOOL,
    CHAR,
    STRING,
    STRING50,
]

def validate(value: Union[int,float,bool,str], dtype: DType) -> bool:
    """Checks that ``value``'s type matches ``dtype``.

    :param value: Value being checked
    :param dtype: DType to check against
    :return: Is ``value`` of type ``DType``?
    """
    return isinstance(value,type(dtype.default))

def get_type_from_value(value: Union[int,float,bool,str]) -> DType:
    """Guess the ``DType`` matching ``value``.

    :param value: Value to find matching dtype
    :return: DType matching ``value``
    """
    if isinstance(value,int):
        return I64
    if isinstance(value,float):
        return F64
    if isinstance(value,bool):
        return BOOL
    if isinstance(value,str):
        return STRING

def get_type_from_string(fmt_str: str) -> DType:
    """Get dtype from ``struct`` format string.

    :param fmt_str: format character with possible
        preceding count.
    :return: Matching dtype for format string
    """
    assert isinstance(fmt_str, str)
    just_char = re.sub(f"[^{valid_chars}]", "", fmt_str)
    assert len(just_char) == 1, f"just_char for '{fmt_str}' is '{just_char}'"
    if len(just_char) == len(fmt_str):
        non_char = None
    else:
        non_char = int(re.sub(r"[a-z]", "", fmt_str))
        assert non_char > 0
    if just_char == I32.value: return I32
    if just_char == I64.value: return I64
    if just_char == F32.value: return F32
    if just_char == F64.value: return F64
    if just_char == BOOL.value: return BOOL
    if just_char == CHAR.value: return CHAR

    if just_char == "s":
        if non_char is None:
            return STRING
        else:
            return STRING[non_char]




import struct
import itertools as it

from . import dtypes
from . import exceptions

from typing import Union, List, Dict, Any


class RowStruct:

    @staticmethod
    def _encode(txt: Union[str,None]) -> bytes:
        """Encode `str` as `bytes` and handles ``None``
        by converting it to an empty `bytes` obj.
        (Assumes ``None``-ness has already been checked
        when writing data row.)

        :param txt: String to be encoded as bytes
        :return: Byte encoding of ``txt``
        """
        if txt is None: return b""
        return txt.encode()

    @staticmethod
    def _decode(txt: bytes) -> str:
        """Decodes bytes object as a string.

        Strips padding null (``\x00``) characters
        and decodes.

        :param txt: Bytes to be decoded
        :return: String decoded from ``txt``
        """
        return txt.strip(b'\x00').decode()

    @staticmethod
    def _getDefault(dtype: Union[str,dtypes.DType]) -> Any:
        """Get the default value for a datatype.

        Used when writing null column values.
        If a null value is being written,
        ``RowStruct`` will set the ``not_null``
        flag to ``False`` and write a default
        value.

        For example, the default values for
        numeric datatypes are ``0``, for booleans
        it's ``False``, and for strings it's
        ``""`` (an empty string).

        :param dtype: Datatype to get default value from
        :return: Default value for ``dtype``
        """
        if isinstance(dtype,dtypes.DType):
            dtype = str(dtype)
        if any(c in dtype for c in "ilfd"):
            return 0
        if "c" in dtype or "s" in dtype:
            return b""
        if "?" in dtype: return False

    def __init__(self, columns: List[str], types: List[dtypes.DType], endian: str = ">"):
        """Wraps the `struct.Struct` class and handles
        null values and strings.

        :param columns: Column names for data rows
        :param types: Column types for ``columns``
        :param endian: Endianness of the data. Options: ``"@=<>!"``. (See python's
            `struct docs <https://docs.python.org/3/library/struct.html#byte-order-size-and-alignment>`_)
        """
        assert len(types) > 0
        assert endian in "@=<>!"
        self.columns = columns
        self.types = types
        self.endian = endian
        self.format = self._makeFmt()
        self.row_struct = struct.Struct(self.format)
        self._strRows = [("s" in str(t)) for t in types]
        self._defaults = [self._getDefault(t) for t in types]

    def _makeFmt(self) -> str:
        """Creates a format string for the `struct.Struct`
        using ``self.types``.

        Adds a boolean flag before each data type
        character to signal if the value is ``NA``.

        Also adds an endian character to the start
        of the format string.

        :return: Format string passed to ``struct.Struct``
        """
        valid_chars = "xc?hilqfds" + "0123456789"
        fmt = self.endian
        fmt += "".join(f"?{t}" for t in self.types)
        for c in fmt[1:]:
            if c not in valid_chars:
                raise exceptions.SchemaError(f"Fmt character '{c}' invalid.")
        assert len(fmt) % 2 == 1, "There should be an even number of chars +1 for `endian`-ness"
        assert len(fmt) > 1
        return fmt

    def _row_dict2list(self, row: Dict[str,Any]) -> list:
        """Convert a column-name-to-row-value dict
        to a ``None``-padded and properly ordered list.

        Used when writing data to a subset of
        columns, or if the columns are out of order
        in the ``row`` dict.

        :param row: Dict mapping from column names to
            column values. ``row`` doesn't need to
            include all of the columns in the table
            but all of ``row``'s keys need to be in
            the table. Also, ``row`` doesn't need to
            include all of the columns in table -- missing
            columns will be filled in with ``None``.

        :return: Correctly orderd list with ``row``'s values,
            ready to be written to the table.
        """
        assert all(r in self.columns for r in row.keys())
        res = {c: None for c in self.columns}
        res.update(row)
        return list(res.values())

    def pack(self, row: Union[List[Any], Dict[str, Any]]) -> bytes:
        """Encodes data from row to a byte string
        that can be written to the table file, per
        the ``RowStruct``'s format string.

        Wrapper around python's ``struct.pack``, which
        handles reordering columns (if necessary -- if
        ``row`` is a dict), as well as NA values.

        :param row: Row of data to be written to table.
        :return: Byte string encoding of ``row``.
        """
        if isinstance(row, dict):
            row = self._row_dict2list(row)
        # Validate the input row
        self._validateTypes(row)
        # Use bool flags to replace NA values
        not_na_flags = [(r is not None) for r in row]
        # Encode the strings, if necessary
        if any(self._strRows):
            row = [(self._encode(r) if is_s else r) for is_s, r
                in zip(self._strRows,row)]
        # Zip and flatten the iterables
        row = it.chain(*[
            (not_na, (val if not_na else dflt))
            for not_na, val, dflt in
            zip(not_na_flags,row,self._defaults)
        ])
        return self.row_struct.pack(*row)

    def _validateTypes(self, row: list):
        """Confirms the types in ``row`` before
        adding them to a table.

        :param row: A list of values to be added to
            a table in the database.
        :raises exceptions.SchemaError: If a value in row
            doesn't match the propper dtype.
        """
        for v, t in zip(row,self.types):
            if not t.validate(v):
                raise exceptions.SchemaError(
                    f'Row value "{v}" is not of type "{t}".')

    def unpack(self, data: bytes) -> List[Any]:
        """Decodes a byte encoding of a row of data
        from the table file.

        Wrapper around python's ``struct.unpack``, which
        handles NA values and strings (which are padded
        in the encoding process).

        :param data: byte encoding of row data
        :return: Row data in list form
        """
        b = self.row_struct.unpack(data)
        assert len(b) > 0
        assert len(b) % 2 == 0
        flags, row = b[::2], b[1::2]
        row = ((self._decode(r) if is_s else r)
            for is_s, r in zip(self._strRows,row))
        row = ((r if f else None)
            for f, r in zip(flags,row))
        return list(row)



"""
"""

import json
import shutil
from pathlib import Path
from datetime import datetime as dt

from . import util
from . import dtypes
from .RowStruct import RowStruct

from typing import Union, Dict, Any, Sequence, List, Callable, Iterable


class Database:
    @staticmethod
    def _validateDirectory(db_path: Union[str,Path]):
        """Check that the directory is valid.

        :param directory: Location of the database directory
        :raises AssertionError: If the directory isn't valid
        """
        db_path = Path(db_path)
        md_file = db_path / "metadata.json"
        assert md_file.exists()
        assert bool(json.loads(md_file.read_text()))

    @classmethod
    def _new(cls, name: str = "db.tdb", path: str = "."):
        """Initialize the filestructure of a new
        database.

        :param filename: filename for new database
        """
        filename = Path(path) / name
        assert not filename.exists(), "`filename` already exists"
        # Make the database directory
        filename.mkdir()
        # And key child directories
        (filename / "tables").mkdir()
        (filename / "indexes").mkdir()
        metafile = filename / "metadata.json"
        metafile.touch()
        metafile.write_text(
            json.dumps({
                "db-name": name,
                "tables": {},
                "created": dt.now().strftime("%Y-%m-%d %H:%M:%S")
        }))

    @classmethod
    def new(cls, name: str = "db.tdb", path: str = "."):
        """Create a new database.

        :param filename: Path to the new databasae
        :return: New Database instance
        """
        cls._new(name,path)
        return cls(name,path)

    def __init__(self, name: str = "db.tdb", path: str = "."):
        """Creates an instance of a `toydb.Database`.

        If a database doesn't already exist, it will
        create a new one by calling ``Database.new()``.

        :param filename: Path to the database directory
        """
        self.filename = Path(path) / name
        if self.filename.exists():
            self._validateDirectory(self.filename)
        else:
            self._new(name,path)
        self.metadata = self._loadMetadata()
        self.name = self.metadata.get("db-name")
        self._structs = self._loadStructs()

    def __str__(self):
        return f"<toydb.Database {self.name}>"

    def __repr__(self):
        return str(self)

    def remove(self):
        """Deletes a database folder and
        all subdirectories."""
        shutil.rmtree(self.filename)

    def listTables(self) -> List[str]:
        """Get a list of Database table names.

        :return: List of DB table names
        """
        return list(self.metadata["tables"].keys())

    def getTableSchema(self, table_name: str) -> Dict[str,dtypes.DType]:
        """Get the schema for table `table_name`.

        :param table_name: Name of existing table in DB
        :return: `dict` mapping from `str` column name to `dtype.DType`
        """
        table_name = table_name.lower()
        assert table_name in self.listTables()
        return self.metadata["tables"][table_name]["schema"]

    def getTableColumns(self, table_name: str) -> List[str]:
        """Get a list of column names in a table.

        :param table_name: Name of existing table in DB
        """
        return list(self.getTableSchema(table_name))

    def createTable(self, table_name: str, schema: Dict[str,dtypes.DType],
        if_not_exists: bool = False):
        """Create a new DB table.

        :param table_name: Name of new table
        :param schema: A dictionary mapping from column
            names to their datatypes. Note: Text datatypes
            need their max lengths specified.
        :param if_not_exists: If ``True`` and the table
            already exists, it won't be overwritten,
            otherwise it will.
        """
        table_name = table_name.lower()
        assert " " not in table_name
        if if_not_exists and table_name in self.listTables():
            return
        filename = self.filename / "tables" / util.md5(table_name)
        filename.touch()
        self.metadata["tables"][table_name] = {
            "schema": schema,
            "indexes": [], # NOTE: Indexes not implemented
            "filename": str(filename)
        }
        self._writeMetadata()
        self._structs = self._loadStructs()

    def _loadMetadata(self) -> dict:
        """Read metadata from file.

        :return: Database metadata `dict`
        """
        mdf = self.filename / "metadata.json"
        assert mdf.exists(), "Metadata file doesn't exist"
        # Read the raw JSON file
        md = json.loads(mdf.read_text())
        # Convert string-dtypes to `DType` objects
        for td in md["tables"].values():
            td["schema"] = {k: dtypes.get_type_from_string(v)
                for k, v in td["schema"].items()}
        return md

    def _writeMetadata(self):
        """Save the current state of the metadata
        to the metadata json file in the datebase
        directory.
        """
        # Write out metadata using custom JSONEncoder
        with (self.filename / "metadata.json").open("w") as f:
            json.dump(self.metadata,f,
                indent=2,cls=dtypes.JSONEncoder)

    def _loadStructs(self) -> Dict[str,dtypes.DType]:
        """Load structs from the metadata file.

        :return: Mapping from tables to `RowStruct`
        """
        return {tn: RowStruct(
            list(d["schema"].keys()),
            list(d["schema"].values()))
            for tn, d in self.metadata["tables"].items()}

    def printSchema(self, table_name: str):
        """Print a table's schema of column
        names and dtypes.

        :param table_name: Existing table in the database.
        """
        table_name = table_name.lower()
        table = self.metadata["tables"].get(table_name)
        assert table is not None, f"Table \"{table_name}\" doesn't exist."
        schema = table.get("schema")
        max_col_len = max(map(len,schema.keys()))
        col_header = f"Table: \"{table_name}\""
        print("","="*int(len(col_header)*1.5))
        print("",col_header)
        print("","="*int(len(col_header)*1.5))
        for col, dtype in schema.items():
            print(f" ... {col:{max_col_len}s} :: {dtype.name}")
        print("","="*int(len(col_header)*1.5))

    def _readLine(self, table_name: str, line_number: int = 0) -> List[Any]:
        """Seek then read a single row
        from a table in the database.

        :param table_name: Table to search
        :param line_number: Line number of row to read
        :return: Row from table as list
        """
        table_name = table_name.lower()
        assert table_name in self.metadata["tables"]
        tablefile = Path(self.metadata["tables"][table_name]["filename"])
        rstruct = self._structs.get(table_name)
        struct_size = rstruct.row_struct.size
        offset = struct_size * line_number
        if offset < 0:
            whence = 2
        with tablefile.open("rb") as f:
            f.seek(offset,whence)
            return rstruct.unpack(f.read(struct_size))

    def _iterReadBytes(self, filename: str, n: int) -> Iterable[bytes]:
        """Generator function for reading a
        binary file `n` bytes at a time.

        :param filename: Filename of binary file
        :param n: Number of bytes to read at a time
        :yields: ``n`` bytes from ``filename``
        """
        with open(filename,"rb") as f:
            while True:
                line = f.read(n)
                if not line: break
                yield line

    def _iterReadAllLines(self, table_name: str) -> Iterable[List]:
        """Generator function for iterating over
        all lines of a database table.

        :param table_name: Name of table in database
        :yields: Row of data from ``table_name``
        """
        table_name = table_name.lower()
        assert table_name in self.metadata["tables"]
        tablefile = Path(self.metadata["tables"][table_name]["filename"])
        lock = Path(str(tablefile) + ".lock")
        assert not lock.exists()
        rstruct = self._structs.get(table_name)
        row_size = rstruct.row_struct.size
        for row in self._iterReadBytes(tablefile,row_size):
            yield rstruct.unpack(row)

    def _iterReadAllDict(self, table_name: str):
        cols = list(self.metadata["tables"][table_name]["schema"].keys())
        for row in self._iterReadAllLines(table_name):
            yield dict(zip(cols,row))

    def _readAllLines(self, table_name: str) -> List[tuple]:
        """Read all lines in a database and return
        it as a list of tupples.

        :param table_name: Table in the database
        :return: All of the data in ``table_name``
        """
        return [tuple(row) for row
            in self._iterReadAllLines(table_name)]

    def query(self, from_: str, select: List[Union[str,Dict[str,Callable]]] = "*", where = None,
        limit: int = None):
        """Query a database using SQL(-ish) syntax.

        :param select: Columns to select
        :param from_: DB table to select from
        :param where: Conditionally filter results with a callable function
        :param limit: Limit the number of results
        """
        table_name = from_.lower()
        assert table_name in self.listTables()
        if select == "*":
            select = self.getTableColumns(table_name)
        itr = self._iterReadAllDict(table_name)
        if limit is not None and limit > 0:
            itr = util.iter_limit(itr,limit)
        # Create SELECT getters
        iden = lambda val: val
        if isinstance(select,str):
            select = {select:iden}
        if isinstance(select,(list,tuple)):
            select = {k: iden for k in select}
        select = {k.lower():v for k, v in select.items()}
        # SELECT and WHERE iterator
        result = (
            tuple(get(row[col]) for col, get in select.items())
            for row in itr if where is None or where(row))
        # Limit the result
        if limit is not None and limit > 0:
            return list(util.iter_limit(result,limit))
        return list(result)

    def insert(self, table_name: str, row: Union[Sequence[Any], Dict[str, Any]]):
        """Add a new row of data into a table.

        :param table_name: Table in the database
        :param row: Row of data to add to table
        """
        table_name = table_name.lower()
        assert table_name in self.metadata["tables"]
        tablefile = Path(self.metadata["tables"][table_name]["filename"])
        rstruct = self._structs.get(table_name)
        data = rstruct.pack(row)
        # Append the data
        with tablefile.open("ab") as f:
            f.write(data)

    def insertMany(self, table_name: str,
        rows: Iterable[Union[Sequence[Any], Dict[str, Any]]]):
        """Insert multiple rows of data into a table.

        :param table_name: Name of table in database
        :param rows: Iterable of rows to add to table
        """
        for row in rows:
            self.insert(table_name,row)

    def _createTempTable(self, table_name: str) -> Path:
        """Create a temporary table version of
        a database table.

        :param table_name: Table in the database
        :return: Path to the new table
        """
        assert table_name in self.metadata["tables"]
        tbl_path = Path(self.metadata["tables"][table_name]["filename"])
        tmp_table = tbl_path.with_suffix(".tmp")
        tmp_table.touch()
        return tmp_table

    def delete(self, table_name: str, where: Callable[[dict],bool]):
        """Delete rows from a table in the
        database where callable ``where``
        evaluates to true.

        Similar to the SQL ``DELETE FROM`` command.

        :param table_name: Table in the database
        :param where: Callable that gets a row from the
            table as an argument (as a dict, mapping
            from column name to value) and should return
            ``True`` if that row should be deleted and
            ``False`` otherwise.
        """
        assert table_name in self.metadata["tables"]
        tbl_path = Path(self.metadata["tables"][table_name]["filename"])
        tmp_path = self._createTempTable(table_name)
        rstruct = self._structs.get(table_name)
        for row in self._iterReadAllDict(table_name):
            if not where(row):
                self.insert(table_name,row)
        # "commit" the change
        tmp_path.rename(tbl_path)

    def dropTable(self, table_name: str):
        """Delete a table from the database.

        :param table_name: Table in database
        """
        assert table_name in self.metadata["tables"]
        table = Path(self.metadata["tables"][table_name]["filename"])
        table.unlink()
        del self.metadata["tables"][table_name]
        self._writeMetadata()




import hashlib
import itertools as it

from typing import Iterable, Generator

def md5(text: str) -> str:
    """md5 hash function.

    Used for creating ``filenames`` based
    on ``table_names``.

    :param text: String to hash
    :return: Hex digest of ``text``
    """
    return hashlib.md5(text.encode()).hexdigest()

def iter_limit(itr: Iterable, limit: int) -> Generator:
    """Generator function that limits the results.

    :param itr: Iterable to limit
    :param limit: Max number of values to yield
        from ``itr``. Must be a positive integer.
    :yields: Up to ``limit`` values from ``itr``.
    """
    for i, v in zip(it.count(),itr):
        if i >= limit: break
        yield v





import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="toydb",
    version="0.1.0",
    author="Austin Poor",
    author_email="austinpoor@gmail.com",
    description="A small, toy database written in pure python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/a-poor/toydb",
    packages=setuptools.find_packages(),
     classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",

        "Topic :: Database",
        "Programming Language :: SQL",
        "Development Status :: 3 - Alpha",
    ],
    python_requires='>=3.6',
)
